# © Copyright Databand.ai, an IBM Company 2022

import pendulum
import pkg_resources

from dbnd._vendor import version


pendulum_version = pkg_resources.get_distribution("pendulum").version
if version.parse(pendulum_version) >= version.parse("2.0.0"):
    pendulum_max_dt = pendulum.DateTime.max
    pendulum_min_dt = pendulum.DateTime.min
else:
    pendulum_max_dt = pendulum.max
    pendulum_min_dt = pendulum.min



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os

from typing import ContextManager, Optional

from airflow.operators.subdag_operator import SubDagOperator
from airflow.utils.log.file_task_handler import FileTaskHandler
from more_itertools import first_true

from dbnd import config, dbnd_bootstrap, dbnd_tracking_stop
from dbnd._core.constants import AD_HOC_DAG_PREFIX
from dbnd._core.context.databand_context import DatabandContext, new_dbnd_context
from dbnd._core.log import dbnd_log_debug, dbnd_log_exception
from dbnd._core.log.buffered_log_manager import BufferedLogManager
from dbnd._core.tracking.airflow_dag_inplace_tracking import calc_task_key_from_af_ti
from dbnd._core.tracking.script_tracking_manager import is_dbnd_tracking_active
from dbnd._core.utils.basics import environ_utils
from dbnd_airflow.tracking.dbnd_airflow_conf import (
    get_dbnd_config_dict_from_airflow_connections,
    get_sync_status_and_tracking_dag_ids_from_dbnd_conf,
)
from dbnd_airflow.tracking.execute_tracking import is_dag_eligable_for_tracking
from dbnd_airflow.tracking.fakes import FakeTaskRun
from dbnd_airflow.utils import get_task_run_attempt_uid_from_af_ti


AIRFLOW_FILE_TASK_HANDLER = FileTaskHandler.__name__

AIRFLOW_TASK_LOGGER = "airflow.task"


class DbndAirflowLogHandler(logging.Handler):
    """
    This is a logger handler that is used as an Entry Point to airflow run.
    It's injected to the Logger(name="airflow.task"), and used by entering the context on the beginning of the task
    instance run, and getting close when the task instance is done.
    """

    def __init__(self, logger):
        logging.Handler.__init__(self)

        self.dbnd_context: Optional[DatabandContext] = None
        self.dbnd_context_manage: Optional[ContextManager[DatabandContext]] = None

        self.task_run_attempt_uid = None
        self.task_env_key = None

        self.airflow_logger = logger

        self.in_memory_log_manager: Optional[BufferedLogManager] = None

    def _set_from_airflow_handler(self):
        airflow_logger = self.airflow_logger
        inheriting_handler = first_true(
            airflow_logger.handlers,
            pred=lambda handler: handler.__class__.__name__
            == AIRFLOW_FILE_TASK_HANDLER,
            default=airflow_logger.handlers[0] if airflow_logger.handlers else None,
        )
        if inheriting_handler:
            handler_formatter = inheriting_handler.formatter
            self.setFormatter(handler_formatter)
            for handler_filter in inheriting_handler.filters:
                self.addFilter(handler_filter)

    def set_context(self, ti):
        try:
            self._set_context(ti)
        except Exception:
            dbnd_log_exception("Failed to initialize DbndAirflowHandler")

    def _set_context(self, ti):
        """
        Airflow's log handler use this method to setup the context when running a TaskInstance(=ti).
        We use this method to setup the dbnd context and communicate information to
        the `<airflow_operator>_execute` task, that we create in `execute_tracking.py`.
        """
        # we setting up only when we are not in our own orchestration dag
        if ti.dag_id.startswith(AD_HOC_DAG_PREFIX):
            return

        dbnd_config_from_connection = get_dbnd_config_dict_from_airflow_connections()
        (
            sync_enabled,
            tracking_list,
        ) = get_sync_status_and_tracking_dag_ids_from_dbnd_conf(
            dbnd_config_from_connection
        )
        if not sync_enabled:
            return

        if not is_dag_eligable_for_tracking(ti.dag_id, tracking_list=tracking_list):
            return

        # we are not tracking SubDagOperator
        if ti.operator is None or ti.operator == SubDagOperator.__name__:
            return

        # make sure we are not polluting the airflow logs
        dbnd_bootstrap()

        if config.getboolean("mlflow_tracking", "databand_tracking"):
            self.airflow_logger.warning(
                "dbnd can't track mlflow and airflow together please disable dbnd config "
                "`databand_tracking` in section `mlflow_tracking`"
            )
            return

        # Airflow is running with two process `--local` and `--raw`.
        # But we want the handler to run only once (Idempotency)
        # So we are using an environment variable to sync those two process
        task_key = calc_task_key_from_af_ti(ti)

        if ti.raw:

            from dbnd._core.settings.tracking_log_config import TrackingLoggingConfig

            logger_config = TrackingLoggingConfig()

            self.in_memory_log_manager = BufferedLogManager(
                max_head_bytes=logger_config.preview_head_bytes,
                max_tail_bytes=logger_config.preview_tail_bytes,
            )
            dbnd_log_debug(
                f"Initiated In Memory Log Manager with task {task_key}: "
                f"head={logger_config.preview_head_bytes} "
                f"tail={logger_config.preview_tail_bytes}, "
                f"context={ self.dbnd_context is not None }. "
            )

        if os.environ.get(task_key):
            # This key is already set which means we are in `--raw run`
            return

        # We are in the outer `run`
        self.task_env_key = task_key
        # marking the environment with the current key for the
        environ_utils.set_on(task_key)

        from dbnd_airflow.tracking.dbnd_airflow_conf import (
            set_dbnd_config_from_airflow_connections,
        )

        # When we are in `--raw run`, in tracking, it runs the main airflow process
        # for every task, which made some of the features to run twice,
        # once when the `worker` process ran, and once when the `main` one ran,
        # which made some of the features to run with different configurations.
        # it still runs twice, but know with the same configurations.
        set_dbnd_config_from_airflow_connections(
            dbnd_config_from_connection=dbnd_config_from_connection
        )

        self.task_run_attempt_uid = get_task_run_attempt_uid_from_af_ti(ti)

        # context with disabled logs
        self.dbnd_context_manage: ContextManager[DatabandContext] = new_dbnd_context()
        self.dbnd_context = self.dbnd_context_manage.__enter__()

    def close(self):
        if self.dbnd_context and self.in_memory_log_manager:
            try:
                fake_task_run = FakeTaskRun(
                    task_run_attempt_uid=self.task_run_attempt_uid
                )
                in_memory_log_body = self.in_memory_log_manager.get_log_body()
                self.dbnd_context.tracking_store.save_task_run_log(
                    task_run=fake_task_run, log_body=in_memory_log_body
                )
                dbnd_log_debug(
                    "Saved Airflow log to Databand service (size=%s)"
                    % len(in_memory_log_body or "")
                )
            except Exception:
                self.airflow_logger.exception("Exception occurred when saving task log")

        self.dbnd_context = None
        self.in_memory_log_manager = None

        try:
            if self.dbnd_context_manage:
                if is_dbnd_tracking_active():
                    # Stops and clears the script tracking if exists and the user forgot to stop it.
                    self.airflow_logger.warning(
                        "A Databand Context from tracking is still active, did you forget to"
                        " use 'dbnd_tracking_stop'?\nWe encourage you to use 'dbnd_tracking' "
                        "context instead of 'dbnd_tracking_start' and 'dbnd_tracking_stop'"
                    )
                    dbnd_tracking_stop()

                self.dbnd_context_manage.__exit__(None, None, None)
        except Exception:
            self.airflow_logger.exception(
                "Exception occurred when trying to exit dbnd_context_manager"
            )
        finally:
            self.dbnd_context_manage = None

        try:
            if self.task_env_key and self.task_env_key in os.environ:
                del os.environ[self.task_env_key]
        except Exception:
            self.airflow_logger.exception(
                "Exception occurred when trying to remove task_env_key from env"
            )
        finally:
            self.task_env_key = None

    def emit(self, record):
        try:
            if self.in_memory_log_manager:
                msg = self.format(record)
                self.in_memory_log_manager.add_log_msg(msg)
        except RecursionError:  # See issue 36272
            # handle errors in log processing
            raise
        except Exception:
            self.handleError(record)


def add_dbnd_log_handler():
    """
    Build and inject the dbnd handler to airflow's logger.
    """
    try:
        airflow_logger = logging.getLogger(AIRFLOW_TASK_LOGGER)
        dbnd_handler = DbndAirflowLogHandler(logger=airflow_logger)
        airflow_logger.addHandler(dbnd_handler)
    except Exception:
        dbnd_log_exception("Failed to add dbnd log handler to airflow logger")



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import json
import logging
import os
import sys

from typing import Dict

from airflow.models import BaseOperator

from dbnd._core.utils.basics.memoized import cached
from dbnd._core.utils.git import get_git_commit, is_git_dirty
from dbnd._core.utils.uid_utils import source_md5
from dbnd.utils.helpers import get_callable_name
from dbnd_airflow.export_plugin.compat import get_task_log_reader


logger = logging.getLogger(__name__)

MAX_LOGS_SIZE_IN_BYTES = 10000
MAX_RECURSIVE_CALL_NUM = 5
TASK_ARG_TYPES = (str, float, bool, int, datetime.datetime)


def resolve_attribute_or_default_value(obj, attribute, default_value):
    if hasattr(obj, attribute):
        return getattr(obj, attribute)
    return default_value


def resolve_attribute_or_default_attribute(obj, attributes_list, default_value=None):
    for attribute in attributes_list:
        if hasattr(obj, attribute):
            return getattr(obj, attribute)
    return default_value


def interval_to_str(schedule_interval):
    if isinstance(schedule_interval, datetime.timedelta):
        if schedule_interval == datetime.timedelta(days=1):
            return "@daily"
        if schedule_interval == datetime.timedelta(hours=1):
            return "@hourly"
    return str(schedule_interval)


def _get_log(ti, task):
    try:
        ti.task = task
        af_logger = logging.getLogger("airflow.task")
        task_log_reader = get_task_log_reader()
        handler = next(
            (
                handler
                for handler in af_logger.handlers
                if handler.name == task_log_reader
            ),
            None,
        )
        logs, metadatas = handler.read(ti, ti._try_number, metadata={})
        if not logs:
            return None
        all_logs = logs[0]
        logs_size = sys.getsizeof(all_logs)
        if logs_size < MAX_LOGS_SIZE_IN_BYTES:
            return all_logs

        result = all_logs[-MAX_LOGS_SIZE_IN_BYTES:] + "... ({} of {})".format(
            MAX_LOGS_SIZE_IN_BYTES, len(all_logs)
        )
        return result
    except Exception:
        pass
    finally:
        del ti.task


@cached()
def _get_git_status(path):
    commit = get_git_commit(path) or ""
    is_dirty = is_git_dirty(path) or False
    return commit, not is_dirty


def _get_source_code(t):
    # type: (BaseOperator) -> str
    # TODO: add other "code" extractions
    # TODO: maybe return it with operator code as well
    try:
        from airflow.operators.bash_operator import BashOperator
        from airflow.operators.python_operator import PythonOperator

        if isinstance(t, PythonOperator):
            import inspect

            return inspect.getsource(t.python_callable)
        elif isinstance(t, BashOperator):
            return t.bash_command
    except Exception:
        pass


def _get_module_code(t):
    # type: (BaseOperator) -> str
    try:
        from airflow.operators.python_operator import PythonOperator

        if isinstance(t, PythonOperator):
            import inspect

            return inspect.getsource(inspect.getmodule(t.python_callable))
    except Exception:
        pass


def _add_source_code(tasks_hash_to_source, source_code):
    source_hash = source_md5(source_code)
    tasks_hash_to_source[source_hash] = source_code
    return source_hash


def _get_command_from_operator(t):
    # type: (BaseOperator) -> str
    from airflow.operators.bash_operator import BashOperator
    from airflow.operators.python_operator import PythonOperator

    if isinstance(t, BashOperator):
        return "bash_command='{bash_command}'".format(bash_command=t.bash_command)
    elif isinstance(t, PythonOperator):
        func_name = get_callable_name(t.python_callable)
        return "python_callable={func}, op_kwargs={kwrags}".format(
            func=func_name, kwrags=t.op_kwargs
        )


def _extract_args_from_dict(t_dict, call_num=1):
    # type: (Dict) -> Dict[str]
    if not t_dict:
        return {}

    if isinstance(t_dict, str) and t_dict.startswith("{"):
        # try load json. is this correct at all?
        try:
            t_dict = json.loads(t_dict)
        except Exception:
            logger.debug("String looked like json but failed to load: %s", t_dict)

    if isinstance(t_dict, str):
        t_dict = {"value": t_dict}

    try:
        # Return only numeric, bool and string attributes
        res = {}
        for k, v in t_dict.items():
            if v is None or isinstance(v, TASK_ARG_TYPES):
                res[k] = v
            elif isinstance(v, list):
                res[k] = [
                    val for val in v if val is None or isinstance(val, TASK_ARG_TYPES)
                ]
            elif isinstance(v, dict):
                if call_num <= MAX_RECURSIVE_CALL_NUM:
                    res[k] = _extract_args_from_dict(v, call_num + 1)
                else:
                    continue
        return res
    except Exception as ex:
        task_id = t_dict.get("task_id") or t_dict.get("_dag_id")
        logging.error("Could not collect task args for %s: %s", task_id, ex)
        return {}


def _read_dag_file(dag_file):
    # TODO: Change implementation when this is done:
    # https://github.com/apache/airflow/pull/7217

    if dag_file and os.path.exists(dag_file):
        with open(dag_file) as file:
            try:
                return file.read()
            except Exception:
                pass

    return ""



# © Copyright Databand.ai, an IBM Company 2022

import typing

from datetime import datetime
from typing import List

import attr

from airflow.models import BaseOperator, DagRun, TaskInstance
from airflow.utils.net import get_hostname

from dbnd._core.utils.uid_utils import source_md5
from dbnd_airflow.export_plugin.helpers import (
    _add_source_code,
    _extract_args_from_dict,
    _get_command_from_operator,
    _get_module_code,
    _get_source_code,
    _read_dag_file,
    interval_to_str,
    resolve_attribute_or_default_attribute,
    resolve_attribute_or_default_value,
)


if typing.TYPE_CHECKING:
    from typing import List

    from airflow.models import DAG, DagModel, DagTag


class ETask(object):
    def __init__(
        self,
        upstream_task_ids=None,
        downstream_task_ids=None,
        task_type=None,
        task_source_code=None,
        task_source_hash=None,
        task_module_code=None,
        module_source_hash=None,
        dag_id=None,
        task_id=None,
        retries=None,
        command=None,
        task_args=None,
    ):
        self.upstream_task_ids = list(upstream_task_ids)  # type: List[str]
        self.downstream_task_ids = list(downstream_task_ids)  # type: List[str]
        self.task_type = task_type
        self.task_source_code = task_source_code
        self.task_source_hash = task_source_hash
        self.task_module_code = task_module_code
        self.module_source_hash = module_source_hash
        self.dag_id = dag_id
        self.task_id = task_id
        self.retries = retries
        self.command = command
        self.task_args = task_args

    @staticmethod
    def from_task(t, include_task_args, dag, include_source=True):
        # type: (BaseOperator, bool, DAG, bool) -> ETask
        module_code = _get_module_code(t) or _read_dag_file(dag.fileloc)
        return ETask(
            upstream_task_ids=t.upstream_task_ids,
            downstream_task_ids=t.downstream_task_ids,
            task_type=t.task_type,
            task_source_hash=source_md5(_get_source_code(t)),
            module_source_hash=source_md5(module_code),
            dag_id=t.dag_id,
            task_id=t.task_id,
            retries=t.retries,
            command=_get_command_from_operator(t),
            task_args=_extract_args_from_dict(vars(t)) if include_task_args else {},
        )

    def as_dict(self):
        return dict(
            upstream_task_ids=self.upstream_task_ids,
            downstream_task_ids=self.downstream_task_ids,
            task_type=self.task_type,
            task_source_code=self.task_source_code,
            task_source_hash=self.task_source_hash,
            task_module_code=self.task_module_code,
            module_source_hash=self.module_source_hash,
            dag_id=self.dag_id,
            task_id=self.task_id,
            retries=self.retries,
            command=self.command,
            task_args=self.task_args,
        )


class EDagRun(object):
    db_fields = [
        "dag_id",
        "id",
        "start_date",
        "state",
        "end_date",
        "execution_date",
        "conf",
        "run_id",
    ]

    @classmethod
    def query_fields(cls):
        return [getattr(DagRun, key) for key in cls.db_fields]

    def __init__(
        self,
        dag_id,
        dagrun_id,
        start_date,
        state,
        end_date,
        execution_date,
        task_args,
        run_id,
    ):
        self.dag_id = dag_id
        self.dagrun_id = dagrun_id
        self.start_date = start_date
        self.state = state
        self.end_date = end_date
        self.execution_date = execution_date
        self.task_args = task_args
        self.run_id = run_id

    @classmethod
    def from_db_fields(
        cls,
        dag_id,
        dagrun_id,
        start_date,
        state,
        end_date,
        execution_date,
        conf,
        run_id,
    ):
        return cls(
            dag_id,
            dagrun_id,
            start_date,
            state,
            end_date,
            execution_date,
            (_extract_args_from_dict(conf) if conf else {}),
            run_id,
        )

    def __hash__(self):
        return hash(self.dagrun_id)

    def __eq__(self, other):
        return isinstance(other, EDagRun) and self.dagrun_id == other.dagrun_id

    def as_dict(self):
        return dict(
            dag_id=self.dag_id,
            dagrun_id=self.dagrun_id,
            start_date=self.start_date,
            state=self.state,
            end_date=self.end_date,
            execution_date=self.execution_date,
            task_args=self.task_args,
            run_id=self.run_id,
        )


class EDag(object):
    def __init__(
        self,
        description,
        root_task_ids,
        tasks,
        owner,
        dag_id,
        schedule_interval,
        catchup,
        start_date,
        end_date,
        dag_folder,
        hostname,
        source_code,
        module_source_hash,
        tasks_hash_to_source,
        is_subdag,
        task_type,
        task_args,
        is_active,
        is_paused,
        git_commit,
        is_committed,
        tags,
    ):
        self.description = description
        self.root_task_ids = root_task_ids  # type: List[str]
        self.tasks = tasks  # type: List[ETask]
        self.tags = tags  # type: List[DagTag]
        self.owner = owner
        self.dag_id = dag_id
        self.schedule_interval = schedule_interval
        self.catchup = catchup
        self.start_date = start_date
        self.end_date = end_date
        self.dag_folder = dag_folder
        self.hostname = hostname
        self.source_code = source_code
        self.module_source_hash = module_source_hash
        self.tasks_hash_to_source = tasks_hash_to_source
        self.is_subdag = is_subdag
        self.task_type = task_type
        self.task_args = task_args
        self.is_active = is_active
        self.is_paused = is_paused
        self.git_commit = git_commit
        self.is_committed = is_committed

    @staticmethod
    def from_dag(
        dag,
        dm,
        dag_folder,
        include_task_args,
        git_commit,
        is_committed,
        raw_data_only=False,
        include_source=True,
    ):
        # type: (DAG, DagModel, str, bool, str, bool, bool, bool) -> EDag
        # Can be Dag from DagBag or from DB, therefore not all attributes may exist
        source_code = _read_dag_file(dag.fileloc)

        tasks_hash_to_source = {}
        if include_source:
            tasks = getattr(dag, "tasks", [])
            for task in tasks:
                _add_source_code(
                    tasks_hash_to_source,
                    _get_module_code(task) or _read_dag_file(dag.fileloc),
                )
                _add_source_code(tasks_hash_to_source, _get_source_code(task))

        return EDag(
            description=dag.description or "",
            root_task_ids=[t.task_id for t in getattr(dag, "roots", [])],
            tasks=[
                ETask.from_task(t, include_task_args, dag, include_source)
                for t in getattr(dag, "tasks", [])
            ]
            if not raw_data_only
            else [],
            owner=resolve_attribute_or_default_attribute(dag, ["owner", "owners"]),
            dag_id=dag.dag_id,
            schedule_interval=interval_to_str(dag.schedule_interval),
            catchup=resolve_attribute_or_default_value(dag, "catchup", False),
            start_date=resolve_attribute_or_default_value(dag, "start_date", None),
            end_date=resolve_attribute_or_default_value(dag, "end_date", None),
            dag_folder=dag_folder,
            hostname=get_hostname(),
            source_code=source_code if not raw_data_only and include_source else "",
            tasks_hash_to_source=tasks_hash_to_source,
            module_source_hash=source_md5(source_code),
            is_subdag=dag.is_subdag,
            tags=getattr(dm, "tags", []),
            task_type="DAG",
            task_args=_extract_args_from_dict(vars(dag)) if include_task_args else {},
            is_active=dm.is_active,
            is_paused=dm.is_paused,
            git_commit=git_commit,
            is_committed=is_committed,
        )

    def as_dict(self):
        return dict(
            description=self.description,
            root_task_ids=self.root_task_ids,
            tasks=[t.as_dict() for t in self.tasks],
            tags=[tag.name for tag in self.tags],
            owner=self.owner,
            dag_id=self.dag_id,
            schedule_interval=self.schedule_interval,
            catchup=self.catchup,
            start_date=self.start_date,
            end_date=self.end_date,
            is_committed=self.is_committed,
            git_commit=self.git_commit,
            dag_folder=self.dag_folder,
            hostname=self.hostname,
            source_code=self.source_code,
            module_source_hash=self.module_source_hash,
            tasks_hash_to_source=self.tasks_hash_to_source,
            is_subdag=self.is_subdag,
            task_type=self.task_type,
            task_args=self.task_args,
        )


@attr.s
class AirflowNewDagRun(object):
    id = attr.ib()  # type: int
    dag_id = attr.ib()  # type: str
    execution_date = attr.ib()  # type: datetime
    state = attr.ib()  # type: str
    is_paused = attr.ib()  # type: bool
    has_updated_task_instances = attr.ib()  # type: bool
    max_log_id = attr.ib()  # type: int
    events = attr.ib()  # type: List[str]

    def as_dict(self):
        return dict(
            id=self.id,
            dag_id=self.dag_id,
            execution_date=self.execution_date,
            state=self.state,
            is_paused=self.is_paused,
            has_updated_task_instances=self.has_updated_task_instances,
            max_log_id=self.max_log_id,
            events=self.events,
        )


class AirflowTaskInstance(object):
    def __init__(
        self, dag_id, task_id, execution_date, state, try_number, start_date, end_date
    ):
        self.execution_date = execution_date
        self.dag_id = dag_id
        self.state = state
        self.try_number = try_number
        self.task_id = task_id
        self.start_date = start_date
        self.end_date = end_date

    db_fields = [
        "dag_id",
        "task_id",
        "execution_date",
        "state",
        "_try_number",
        "start_date",
        "end_date",
    ]

    @classmethod
    def query_fields(cls):
        return [getattr(TaskInstance, key) for key in cls.db_fields]

    def as_dict(self):
        return dict(
            dag_id=self.dag_id,
            task_id=self.task_id,
            execution_date=self.execution_date,
            state=self.state,
            try_number=self.try_number,
            start_date=self.start_date,
            end_date=self.end_date,
        )


@attr.s
class AirflowExportMeta(object):
    airflow_version = attr.ib(default=None)  # type: str
    plugin_version = attr.ib(default=None)  # type: str
    airflow_instance_uid = attr.ib(default=None)  # type: str
    api_mode = attr.ib(default=None)  # type: str
    request_args = attr.ib(default=None)  # type: dict
    metrics = attr.ib(default=None)  # type: dict

    def as_dict(self):
        return dict(
            airflow_version=self.airflow_version,
            plugin_version=self.plugin_version,
            airflow_instance_uid=self.airflow_instance_uid,
            api_mode=self.api_mode,
            request_args=self.request_args,
            metrics=self.metrics,
        )


@attr.s
class AirflowExportData(object):
    airflow_export_meta = attr.ib(default=None)  # type: AirflowExportMeta
    error_message = attr.ib(default=None)  # type: str

    def as_dict(self):
        return dict(
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class LastSeenData(AirflowExportData):
    last_seen_dag_run_id = attr.ib(default=None)  # type: int
    last_seen_log_id = attr.ib(default=None)  # type: int

    def as_dict(self):
        return dict(
            last_seen_dag_run_id=self.last_seen_dag_run_id,
            last_seen_log_id=self.last_seen_log_id,
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class NewRunsData(AirflowExportData):
    new_dag_runs = attr.ib(default=None)  # type: List[AirflowNewDagRun]
    last_seen_dag_run_id = attr.ib(default=None)  # type: int
    last_seen_log_id = attr.ib(default=None)  # type: int

    def as_dict(self):
        return dict(
            new_dag_runs=[new_dag_run.as_dict() for new_dag_run in self.new_dag_runs],
            last_seen_dag_run_id=self.last_seen_dag_run_id,
            last_seen_log_id=self.last_seen_log_id,
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class FullRunsData(AirflowExportData):
    task_instances = attr.ib(default=None)  # type: List[AirflowTaskInstance]
    dag_runs = attr.ib(default=None)  # type: List[EDagRun]
    dags = attr.ib(default=None)  # type: List[EDag]

    def as_dict(self):
        return dict(
            task_instances=[
                task_instance.as_dict() for task_instance in self.task_instances
            ],
            dag_runs=[run.as_dict() for run in self.dag_runs],
            dags=[dag.as_dict() for dag in self.dags],
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class DagRunsStatesData(AirflowExportData):
    task_instances = attr.ib(default=None)  # type: List[AirflowTaskInstance]
    dag_runs = attr.ib(default=None)  # type: List[EDagRun]

    def as_dict(self):
        return dict(
            task_instances=[
                task_instance.as_dict() for task_instance in self.task_instances
            ],
            dag_runs=[run.as_dict() for run in self.dag_runs],
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import json

import airflow

from flask import Response

from dbnd._vendor import version


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        # convert dates and numpy objects in a json serializable format
        if isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def json_response(obj):
    return Response(
        response=json.dumps(obj, indent=4, cls=JsonEncoder),
        status=200,
        mimetype="application/json",
    )


# we can not use `dbnd_airflow.compat` here,
# as we have .compat usage at "import" level,
# meaning we might have circular dependency while loading airflow plugins
# (import dbnd_airflow.compat -> import airflow -> import plugin -> import dbnd_airflow.compat again)
AIRFLOW_VERSION_2 = version.parse(airflow.version.version) >= version.parse("2.0.0")
AIRFLOW_VERSION_BEFORE_2_2 = version.parse(airflow.version.version) < version.parse(
    "2.2.0"
)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from contextlib import contextmanager
from itertools import islice
from typing import Any, Dict, Optional

from dbnd import config, dbnd_bootstrap, log_metric, log_metrics
from dbnd._core.configuration.environ_config import is_dbnd_disabled
from dbnd._core.constants import TaskRunState
from dbnd._core.log.dbnd_log import dbnd_log_debug, dbnd_log_exception
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.tracking.airflow_task_context import AirflowTaskContext
from dbnd._core.tracking.managers.callable_tracking import _log_result
from dbnd._core.tracking.script_tracking_manager import (
    dbnd_tracking_stop,
    tracking_start_base,
)
from dbnd._core.utils.basics.environ_utils import env
from dbnd._core.utils.basics.nested_context import safe_nested
from dbnd_airflow.raw_constants import MONITOR_DAG_NAME
from dbnd_airflow.tracking.config import AirflowTrackingConfig
from dbnd_airflow.tracking.dbnd_airflow_conf import (
    AIRFLOW_MONITOR_CONFIG_NAME,
    IS_SYNC_ENABLED_TRACKING_CONFIG_NAME,
    get_dbnd_config_dict_from_airflow_connections,
    get_sync_status_and_tracking_dag_ids_from_dbnd_conf,
    get_tracking_information,
    get_xcoms,
    set_dbnd_config_from_airflow_connections,
)
from dbnd_airflow.tracking.wrap_operators import (
    get_airflow_operator_handlers_config,
    wrap_operator_with_tracking_info,
)
from dbnd_airflow.utils import get_airflow_instance_uid


logger = logging.getLogger(__name__)


def is_monitor_dag(dag_id):
    return dag_id == MONITOR_DAG_NAME


def is_subdag_eligible_for_tracking(dag_id, tracking_list):
    for tracked_dag_id in tracking_list:
        if dag_id.startswith(f"{tracked_dag_id}."):
            return True

    return False


def is_dag_eligable_for_tracking(dag_id, tracking_list):
    # If return value is None, we track all dags. If it's empty list - we don't track anything.
    return (
        tracking_list is None
        or dag_id in tracking_list
        or is_subdag_eligible_for_tracking(dag_id, tracking_list)
        or is_monitor_dag(dag_id)
    )


def _try_to_start_airflow_operator_tracking(context):
    if is_dbnd_disabled():
        dbnd_log_debug("Databand is disabled via project config.")
        # we are not tracking if dbnd is disabled
        return None

    copied_operator = context["task_instance"].task
    # import pydevd_pycharm
    # pydevd_pycharm.settrace('localhost', port=8787, stdoutToServer=True, stderrToServer=True)

    dbnd_log_debug(
        "Running tracked .execute for %s (%s)",
        copied_operator.task_id,
        f"{copied_operator.__class__.__module__}.{copied_operator.__class__.__qualname__}",
    )

    dbnd_config_from_connection = get_dbnd_config_dict_from_airflow_connections()
    sync_enabled, tracking_list = get_sync_status_and_tracking_dag_ids_from_dbnd_conf(
        dbnd_config_from_connection
    )
    if not sync_enabled:
        dbnd_log_debug(
            f"DBND Tracking is disabled via Airflow connection`dbnd_config`"
            f" (at {AIRFLOW_MONITOR_CONFIG_NAME}.{IS_SYNC_ENABLED_TRACKING_CONFIG_NAME} ("
        )
        return None

    if not is_dag_eligable_for_tracking(
        context["task_instance"].dag_id, tracking_list=tracking_list
    ):
        dbnd_log_debug("DAG id is not tracked %s", context["task_instance"].dag_id)
        return None

    # Only here we kinda know that we will start tracking

    dbnd_bootstrap()

    # Set Databand config from Extra section in Airflow dbnd_config connection.
    set_dbnd_config_from_airflow_connections(
        dbnd_config_from_connection=dbnd_config_from_connection
    )

    task_context = extract_airflow_task_context(context)
    # start operator execute run with current airflow context
    task_run: Optional[TaskRun] = tracking_start_base(airflow_context=task_context)
    return task_run


def execute_with_dbnd_tracking(context):
    """
    This function replaces the operator's original `execute` function
    """

    # IMPORTANT!!: copied_operator:
    # ---------------------------------------
    # The task (=operator) is copied when airflow enters to TaskInstance._run_raw_task.
    # Then, only the copy_task (=copy_operator) is changed or called (render jinja, signal_handler,
    # pre_execute, execute, etc..).
    copied_operator = context["task_instance"].task

    dbnd_tracking_task_run = False
    try:
        dbnd_tracking_task_run = _try_to_start_airflow_operator_tracking(context)
    except Exception as e:
        dbnd_log_exception(
            "Exception caught while starting DBND tracking: {}".format(e)
        )

    if not dbnd_tracking_task_run:

        execute = get_execute_function(copied_operator)
        result = execute(copied_operator, context)
        return result

    from airflow.exceptions import AirflowRescheduleException

    # running the operator's original execute function
    try:
        operator_config = getattr(copied_operator, "dbnd_config", {})
        with config(operator_config, "airflow_operator_definition"):
            with af_tracking_context(dbnd_tracking_task_run, context, copied_operator):
                execute = get_execute_function(copied_operator)
                result = execute(copied_operator, context)

    # Check if this is sensor task that is retrying - normal behavior and not really an exception
    except AirflowRescheduleException:
        dbnd_tracking_stop(finalize_run=False)
        raise
    # catch if the original execute failed
    except Exception as ex:
        try:
            error = TaskRunError.build_from_ex(ex, dbnd_tracking_task_run)
            dbnd_tracking_task_run.set_task_run_state(
                state=TaskRunState.FAILED, error=error
            )
        except Exception:
            dbnd_log_exception("Failed to save failed state for the task")

        dbnd_tracking_stop()
        raise

    # if we have a task run here we want to log results and xcoms
    if dbnd_tracking_task_run:
        try:
            track_config = AirflowTrackingConfig.from_databand_context()
            if track_config.track_xcom_values:
                # reporting xcoms as metrix of the task
                log_xcom(context, track_config)

            if track_config.track_airflow_execute_result:
                # reporting the result
                log_operator_result(
                    dbnd_tracking_task_run,
                    result,
                    copied_operator,
                    track_config.track_xcom_values,
                )

        except Exception as e:
            dbnd_log_exception(
                "exception caught while logging airflow operator results {}".format(e)
            )

    # make sure we close and return the original results
    dbnd_tracking_stop()
    return result


def log_xcom(context, track_config):
    task_instance = context["task_instance"]
    xcoms = get_xcoms(task_instance)
    if xcoms:
        # get only the first xcoms
        xcoms_head = islice(xcoms, track_config.max_xcom_length)
        log_metrics(dict(xcoms_head))


def log_operator_result(task_run, result, operator, track_xcom):
    _log_result(task_run, result)

    # after airflow runs the operator it xcom_push the result, so we log it
    if track_xcom and operator.do_xcom_push and result is not None:
        from airflow.models import XCOM_RETURN_KEY

        log_metric(key=XCOM_RETURN_KEY, value=result)


@contextmanager
def af_tracking_context(task_run, airflow_context, operator):
    """
    Wrap the execution with handling the environment management
    """
    if not task_run:
        # aborting -  can't enter the context without task_run_executor
        yield
        return

    try:
        tracking_info = get_tracking_information(airflow_context, task_run)
        airflow_operator_handlers = get_airflow_operator_handlers_config(
            task_run.run.context.settings.tracking.airflow_operator_handlers
        )
        operator_wrapper = wrap_operator_with_tracking_info(
            tracking_info, operator, airflow_operator_handlers
        )

    except Exception as e:
        logger.error(
            "exception caught adding tracking context to operator execution {}"
            "continue without tracking context".format(e),
            exc_info=True,
        )
        yield
        return

    # wrap the execution with tracking info in the environment
    with safe_nested(env(**tracking_info), operator_wrapper):
        yield


def execute_with_dbnd_tracking_for_class(self, context):
    """
    Different wrapper for operator's class
    """
    return execute_with_dbnd_tracking(context)


def track_operator(operator):
    """
    Replace the operator's execute method with our `tracked execute` method
    """
    import inspect

    if inspect.isclass(operator):
        from airflow.models import BaseOperator

        # we only track operators which base on Airflow BaseOperator
        if not issubclass(operator, BaseOperator):
            return operator

        # the operator class is already tracked
        if (
            hasattr(operator, "_tracked_class")
            and operator.__name__ == operator._tracked_class
        ):
            return operator

        # this is the first time we encounter this class so we mark it
        operator._tracked_class = operator.__name__
        operator.__execute__ = operator.execute
        operator.execute = execute_with_dbnd_tracking_for_class
        return operator

    else:
        # the operator instance is already tracked
        if hasattr(operator, "_tracked_instance"):
            return operator

        # this is the first time we encounter this instance so we mark it
        operator._tracked_instance = True

        # if the operator's class is tracked, we can't used `operator.__class__.execute`
        # need to use the original execute (__execute__)
        if (
            hasattr(operator, "_tracked_class")
            and operator.__class__.__name__ == operator._tracked_class
        ):
            operator.__execute__ = operator.__class__.__execute__
        else:
            # this can be a problem when the operator class doesn't implement it's own execute
            operator.__execute__ = operator.__class__.execute

        operator.execute = execute_with_dbnd_tracking

    return operator


def get_execute_function(copied_operator):
    """
    After tracking operator we need to access the original execution function
    """
    if hasattr(copied_operator, "_tracked_instance"):
        return copied_operator.__execute__

    if hasattr(copied_operator, "_tracked_class"):
        return copied_operator.__class__.__execute__

    raise AttributeError("tracked failed to run original operator.execute")


def extract_airflow_task_context(
    airflow_context: Dict[str, Any]
) -> Optional[AirflowTaskContext]:
    """Create AirflowTaskContext for airflow_context dict"""

    task_instance = airflow_context.get("task_instance")
    if task_instance is None:
        return None

    dag_id = task_instance.dag_id
    task_id = task_instance.task_id
    execution_date = str(task_instance.execution_date)
    try_number = task_instance.try_number

    if dag_id and task_id and execution_date:
        return AirflowTaskContext(
            dag_id=dag_id,
            execution_date=execution_date,
            task_id=task_id,
            try_number=try_number,
            context=airflow_context,
            airflow_instance_uid=get_airflow_instance_uid(),
            is_subdag=airflow_context.get("dag").is_subdag,
        )

    logger.debug(
        "airflow context from inspect, at least one of those params is missing"
        "dag_id: {}, execution_date: {}, task_id: {}".format(
            dag_id, execution_date, task_id
        )
    )
    return None



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from collections import OrderedDict
from contextlib import contextmanager
from typing import Any, ContextManager, Dict, Optional

from dbnd._core.log import dbnd_log_debug
from dbnd._core.utils.basics.load_python_module import load_python_callable
from dbnd_airflow.tracking.dbnd_spark_conf import (
    track_databricks_submit_run_operator,
    track_dataproc_pyspark_operator,
    track_dataproc_submit_job_operator,
    track_ecs_operator,
    track_emr_add_steps_operator,
    track_spark_submit_operator,
)


if typing.TYPE_CHECKING:
    from airflow.models.baseoperator import BaseOperator

logger = logging.getLogger(__name__)

# registering operators names to the relevant tracking method
_EXECUTE_TRACKING = OrderedDict(
    [
        # SPARK SUBMIT
        # Airflow 1.x
        (
            "airflow.contrib.operators.spark_submit_operator.SparkSubmitOperator",
            track_spark_submit_operator,
        ),
        # Airflow 2.x
        (
            "airflow.providers.apache.spark.operators.spark_submit.SparkSubmitOperator",
            track_spark_submit_operator,
        ),
        # EMR
        # Airflow 1.x
        (
            "airflow.contrib.operators.emr_add_steps_operator.EmrAddStepsOperator",
            track_emr_add_steps_operator,
        ),
        # Airflow 2.0
        (
            "airflow.providers.amazon.aws.operators.emr_add_steps.EmrAddStepsOperator",
            track_emr_add_steps_operator,
        ),
        (
            "airflow.providers.amazon.aws.operators.emr_pyspark.EmrPySparkOperator",
            track_emr_add_steps_operator,
        ),
        # Airflow 2.2+
        (
            "airflow.providers.amazon.aws.operators.emr.EmrAddStepsOperator",
            track_emr_add_steps_operator,
        ),
        # Databricks - track_databricks_submit_run_operator
        # Airflow 1.x
        (
            "airflow.contrib.operators.databricks_operator.DatabricksSubmitRunOperator",
            track_databricks_submit_run_operator,
        ),
        # Airflow 2.x
        (
            "airflow.providers.databricks.operators.databricks.DatabricksSubmitRunOperator",
            track_databricks_submit_run_operator,
        ),
        # Dataproc - track_dataproc_submit_job_operator
        # Airflow 1.x
        (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitJobOperator",
            track_dataproc_submit_job_operator,
        ),
        # Airflow 2.x
        (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator",
            track_dataproc_pyspark_operator,
        ),
        # ECS
        # Airflow 1.x -
        ("airflow.contrib.operators.ecs_operator.ECSOperator", track_ecs_operator),
        # # Airflow 2.x
        ("airflow.providers.amazon.aws.operators.ecs.ECSOperator", track_ecs_operator),
        # Airflow 2.x - with lowercase Ecs (not ECS)
        ("airflow.providers.amazon.aws.operators.ecs.EcsOperator", track_ecs_operator),
    ]
)


def register_airflow_operator_handler(operator, airflow_operator_handler):
    global _EXECUTE_TRACKING
    if isinstance(operator, type):
        operator = f"{operator.__module__}.{operator.__qualname__}"

    dbnd_log_debug(
        "Registering operator handler %s with %s", operator, airflow_operator_handler
    )
    _EXECUTE_TRACKING[operator] = airflow_operator_handler


def get_airflow_operator_handlers_config(user_config_airflow_operator_handlers=None):
    if user_config_airflow_operator_handlers:
        target = _EXECUTE_TRACKING.copy()
        target.update(user_config_airflow_operator_handlers)
        return target

    return _EXECUTE_TRACKING


def wrap_operator_with_tracking_info(
    tracking_info: Dict[str, str],
    operator: "BaseOperator",
    airflow_operator_handlers: Dict[str, Any],
) -> Optional[ContextManager]:
    """
    Wrap the operator with relevant tracking method, if found such method.
    """

    dbnd_log_debug(
        "Checking for the relevant wrapper for operator '%s' (%s)",
        operator,
        f"{operator.__class__.__module__}.{operator.__class__.__qualname__}",
    )
    for cls in operator.__class__.mro():
        # for user operators we support only FQN matching
        tracking_wrapper = _get_loaded_tracking_wrapper(
            airflow_operator_handlers, cls.__qualname__
        ) or _get_loaded_tracking_wrapper(
            airflow_operator_handlers, f"{cls.__module__}.{cls.__qualname__}"
        )
        if tracking_wrapper:
            dbnd_log_debug(
                "Applying airflow operator wrapper %s at %s",
                operator.task_id,
                tracking_wrapper,
            )
            return tracking_wrapper(operator, tracking_info)
    dbnd_log_debug("There is no specific wrapper for '%s'", operator)
    return None


def _get_loaded_tracking_wrapper(airflow_operator_handlers, name):
    if name not in airflow_operator_handlers:
        # not found
        return None

    tracking_wrapper = airflow_operator_handlers.get(name)
    if not tracking_wrapper:
        # found but empty -> user has disabled the operator
        return None
    if isinstance(tracking_wrapper, str):
        tracking_wrapper = load_python_callable(tracking_wrapper)
        # Update the cache with the loaded function
        airflow_operator_handlers[name] = tracking_wrapper
    return tracking_wrapper


@contextmanager
def track_nope_wrapper(operator, tracking_info):
    yield



# © Copyright Databand.ai, an IBM Company 2022
import logging
import os

import dbnd

from dbnd._core.log import dbnd_log_debug, is_verbose
from dbnd_airflow.tracking.airflow_patching import (
    add_tracking_to_policy,
    patch_airflow_context_vars,
    patch_snowflake_hook,
)
from dbnd_airflow.tracking.dbnd_airflow_handler import (
    AIRFLOW_TASK_LOGGER,
    add_dbnd_log_handler,
)


_airflow_bootstrap_applied = False


def dbnd_airflow_tracking_entrypoint():
    """
    all relevant patches for airflow execution
    """
    global _airflow_bootstrap_applied
    if _airflow_bootstrap_applied:
        return
    _airflow_bootstrap_applied = True  # prevent recursive call/repeated error call

    airflow_logger = logging.getLogger(AIRFLOW_TASK_LOGGER)
    dbnd_log_debug("Airflow wrapping via entrypoint")
    # This code will run every time Airflow imports our plugin
    # it will change current airflow policy to include Databand policy via add_tracking_to_policy
    # -> `dbnd_airflow.tracking.dbnd_dag_tracking.track_task` will be applied on every task
    # -> `dbnd_airflow.tracking.execute_tracking.new_execute` will wrap Operator.execute
    # new_execute -> `dbnd_airflow.tracking.execute_tracking.af_tracking_context` will modify
    #     airflow operator to have "our tracking parameters"
    # new_execute -> `wrap_operator_with_tracking_info` will modify operator to have proper
    #         "ENV/params/spark.configs/jars.." so the internal part is tracked

    # Add policy, that will wrap all Operator.execute() with DBND tracking wrapper
    add_tracking_to_policy()

    # Adds extra DBND_ variables to vars
    patch_airflow_context_vars()

    # Add handler Handler to send logs to DBND
    add_dbnd_log_handler()

    # patch snoflake if available
    patch_snowflake_hook()

    # tracking msg
    if is_verbose():
        dbnd_log_debug("Airflow wrapping has finished (pid=%s)" % os.getpid())

    airflow_logger.info(
        "DBND Tracking Entrypoint {version}".format(version=dbnd.__version__)
    )



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import typing

from contextlib import contextmanager
from typing import Any, Dict, List

import six

from dbnd import dbnd_context
from dbnd._core.configuration.environ_config import (
    ENV_DBND__ENABLE__SPARK_CONTEXT_ENV,
    ENV_DBND_SCRIPT_NAME,
)
from dbnd._core.constants import TaskRunState, UpdateSource
from dbnd._core.log import dbnd_log_exception, is_verbose
from dbnd._core.log.dbnd_log import ENV_DBND__VERBOSE, dbnd_log_debug, dbnd_log_info
from dbnd._core.tracking.airflow_dag_inplace_tracking import (
    get_task_run_uid_for_inline_script,
)
from dbnd._core.tracking.backends import TrackingStore
from dbnd_airflow.tracking.config import TrackingSparkConfig
from dbnd_airflow.tracking.fakes import FakeRun, FakeTask, FakeTaskRun


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from airflow.contrib.operators.databricks_operator import (
        DatabricksHook,
        DatabricksSubmitRunOperator,
    )
    from airflow.contrib.operators.ecs_operator import ECSOperator
    from airflow.contrib.operators.spark_submit_operator import SparkSubmitOperator

logger = logging.getLogger(__name__)


def convert_spark_conf_to_cli_args(conf):
    """Flats a configuration iterable to a list of commands ready to concat to"""
    results = []
    for field, value in six.iteritems(conf):
        results.extend(["--conf", field + "=" + value])
    return results


def _merge_jars(jar_uris, spark_jars):
    # type: (List[str], str) -> List[str]
    """
    Merge jar list, but for jar_file_uris property. spark_jars is comma-separated list of jars
    """
    if spark_jars:
        return jar_uris + spark_jars.split(",")
    return jar_uris


def _normalize_python_script_name(raw_script_path: str) -> typing.Optional[str]:
    if not raw_script_path:
        return None
    try:
        script_name = os.path.basename(raw_script_path)
        if not script_name:
            logger.warning(
                "Unable to determine script name from path %s", raw_script_path
            )
            return None

        return script_name
    except Exception as exc:
        logger.error(
            "Unable to determine script name from path %s, exception: %s",
            raw_script_path,
            exc,
        )
        return None


def get_dbnd_context_spark_conf(
    tracking_info: Dict[str, str], spark_conf: Dict[str, str] = None
) -> Dict[str, str]:
    """
    Takes dbnd context as dict (airflow context + databand task context), builds spark.env properties,
    and enriches it with properties required for proper tracking in client and cluster modes
    """

    #     if try_get_current_task_run():
    #         tracking_info = get_tracking_information(context, try_get_current_task_run())

    dbnd_spark_conf = {
        "spark.env." + key: value for key, value in tracking_info.items()
    }
    dbnd_spark_conf.update(
        {
            # Spark properties to explicitly enable tracking
            "spark.env.DBND__TRACKING": "True",
            # Spark properties to explicitly enable tracking in cluster mode
            "spark.yarn.appMasterEnv.DBND__TRACKING": "True",
            # might be missing in server mode
            "spark.yarn.appMasterEnv.SPARK_ENV_LOADED": "1",
            "spark.yarn.appMasterEnv.DBND__ENABLE__SPARK_CONTEXT_ENV": "True",
            "spark.yarn.appMasterEnv.DBND_HOME": "/tmp/dbnd",
            # Spark properties to explicitly enable tracking in K8S mode
            "spark.kubernetes.driverEnv.DBND__TRACKING": "True",
            "spark.kubernetes.driverEnv.DBND__ENABLE__SPARK_CONTEXT_ENV": "True",
        }
    )

    if is_verbose():
        dbnd_spark_conf.update(
            {
                "spark.env.DBND__VERBOSE": "True",
                "spark.yarn.appMasterEnv.DBND__VERBOSE": "True",
                "spark.kubernetes.driverEnv.DBND__VERBOSE": "True",
            }
        )

    dbnd_spark_conf.update(TrackingSparkConfig.from_databand_context().spark_conf())

    # simple case, used doesn't have his own configuration
    if not spark_conf:
        dbnd_log_debug("Injecting Env variables %s", dbnd_spark_conf)
        return dbnd_spark_conf

    # we need to merge
    merged_conf = spark_conf.copy()
    merged_conf.update(dbnd_spark_conf)

    # Operator spark conf can contain "spark.jars", "spark.driver.extraJavaOptions"
    # and "spark.sql.queryExecutionListeners".
    # To preserve existing properties we need to concat them with patched properties
    for property_name in [
        "spark.driver.extraJavaOptions",
        "spark.jars",
        "spark.sql.queryExecutionListeners",
    ]:
        if property_name in spark_conf and property_name in dbnd_spark_conf:
            merged_conf[property_name] = (
                spark_conf[property_name] + "," + dbnd_spark_conf[property_name]
            )

    dbnd_log_debug("Injecting Env variables with merge: %s", dbnd_spark_conf)
    return merged_conf


def get_dbnd_context_env_vars(tracking_info=None, env_vars=None, script_name=None):
    env_vars_with_dbnd = env_vars.copy() if env_vars else dict()
    env_vars_with_dbnd.update(tracking_info)
    # env_vars_with_dbnd.update(get_databand_url_conf())
    if script_name:
        script_name = _normalize_python_script_name(script_name)
        if script_name:
            env_vars_with_dbnd[ENV_DBND_SCRIPT_NAME] = script_name

    if is_verbose():
        env_vars_with_dbnd[ENV_DBND__VERBOSE] = "True"
    env_vars_with_dbnd[ENV_DBND__ENABLE__SPARK_CONTEXT_ENV] = "True"
    return env_vars_with_dbnd


def get_spark_submit_cli_with_dbnd_context(
    original_command_as_list, dbnd_context_as_cli=None
):
    """This functions augments spark-submit command with dbnd tracking metadata aka dbnd_context_as_cli. If context
    is not provided, a default Airflow templates are set for DAG Id, Task Id, and execution date

    Adds 3 configuration properties to spark-submit to associate spark run with airflow DAG that initiated this run.
    These properties are
        spark.env.AIRFLOW_CTX_DAG_ID  -  name of the Airflow DAG to associate a run with
        spark.env.AIRFLOW_CTX_EXECUTION_DATE - execution_date to associate a run with
        spark.env.AIRFLOW_CTX_TASK_ID" - name of the Airflow Task to associate a run with

    Example:
        >>> get_spark_submit_cli_with_dbnd_context(["spark-submit","my_script.py","my_param"])
        ['spark-submit', '--conf', 'spark.env.AIRFLOW_CTX_DAG_ID={{dag.dag_id}}', '--conf', 'spark.env.AIRFLOW_CTX_EXECUTION_DATE={{ts}}', '--conf', 'spark.env.AIRFLOW_CTX_TASK_ID={{task.task_id}}', 'my_script.py', 'my_param']

    Args:
        original_command_as_list: spark-submit command line as a list of strings.
        dbnd_context_as_cli: optional dbnd_context provided by a user. I

    Returns:
        An updated spark-submit command including dbnd context.

    """
    index = next(
        (
            original_command_as_list.index(elm)
            for elm in original_command_as_list
            if "spark-submit" in elm
        ),
        -1,
    )
    if index == -1:
        dbnd_log_exception(
            "Failed to find 'spark-submit' cmd in %s"
            % " ".join(original_command_as_list)
        )
        return original_command_as_list

    return (
        original_command_as_list[0 : index + 1]
        + dbnd_context_as_cli
        + original_command_as_list[index + 1 :]
    )


@contextmanager
def track_emr_add_steps_operator(operator, tracking_info):
    dbnd_log_info("Updating HadoopJarStep.Args with dbnd context")

    spark_envs = get_dbnd_context_spark_conf(tracking_info)
    # EMR uses list arguments, it's safe to update it

    dbnd_context_as_cli = convert_spark_conf_to_cli_args(spark_envs)
    for step in operator.steps:
        args = step["HadoopJarStep"]["Args"]
        if args and "spark-submit" in args[0]:
            # Add dbnd and airflow context
            step["HadoopJarStep"]["Args"] = get_spark_submit_cli_with_dbnd_context(
                args, dbnd_context_as_cli=dbnd_context_as_cli
            )

        else:
            dbnd_log_exception("spark-submit has been not found in the operator")
    yield


@contextmanager
def track_pyspark_operator_with_spark_args_and_jars_attrs(operator, tracking_info):
    if operator.spark_args is None:
        operator.spark_args = []

    spark_args_array = convert_spark_conf_to_cli_args(
        get_dbnd_context_spark_conf(tracking_info)
    )
    operator.spark_args.extend(spark_args_array)
    yield


@contextmanager
def track_databricks_submit_run_operator(operator, tracking_info):
    # type: (DatabricksSubmitRunOperator, Dict[str, str])-> None
    config = operator.json
    script_name = None  # type: str
    # passing env variables is only supported in new clusters
    cluster = None

    if "new_cluster" in config:
        cluster = config["new_cluster"]
        cluster.setdefault("spark_env_vars", {})
        # via Environment
        cluster["spark_env_vars"] = get_dbnd_context_env_vars(
            tracking_info=tracking_info,
            script_name=config.get("spark_python_task", {}).get("python_file"),
            env_vars=cluster.get("spark_env_vars"),
        )

        # via spark.conf
        cluster["spark_conf"] = get_dbnd_context_spark_conf(
            tracking_info=tracking_info, spark_conf=cluster.get("spark_conf")
        )

        if "spark_python_task" in config:
            script_name = cluster["spark_env_vars"].get(ENV_DBND_SCRIPT_NAME)

    yield
    if script_name:
        # calculate deterministic task uids so we can use it for manual completion
        (
            task_id,
            task_run_uid,
            task_run_attempt_uid,
        ) = get_task_run_uid_for_inline_script(tracking_info, script_name)

        # When dbnd is running inside Databricks, the script can be SIGTERM'd before dbnd will send state to tracker
        # So we need to check whenever the run was succeeded afterwards and set run state manually
        tracking_store = dbnd_context().tracking_store  # type: TrackingStore
        run = FakeRun(source=UpdateSource.airflow_tracking)
        task_run = FakeTaskRun(
            task_run_uid=task_run_uid,
            task_run_attempt_uid=task_run_attempt_uid,
            run=run,
            task=FakeTask(task_name=task_id, task_id=task_id),
            task_af_id=task_id,
        )
        try:
            hook = None
            if hasattr(operator, "_get_hook"):
                hook = operator._get_hook()  # type: DatabricksHook
            elif hasattr(operator, "get_hook"):
                hook = operator.get_hook()  # type: DatabricksHook
            if hook:
                run_page_url = hook.get_run_page_url(operator.run_id)
                state = hook.get_run_state(operator.run_id)

            tracking_store = dbnd_context().tracking_store  # type: TrackingStore
            save_external_links_safe(
                tracking_store, task_run, {"databricks": run_page_url}
            )

            if state.is_successful:
                set_task_run_state_safe(tracking_store, task_run, TaskRunState.SUCCESS)
            else:
                # TODO: error should be extracted from plain Databricks logs
                set_task_run_state_safe(tracking_store, task_run, TaskRunState.FAILED)
        except Exception as exc:
            logger.error(
                "Unable to get inline script run state from Databricks. Setting task run state to Failed: %s",
                exc,
            )
            set_task_run_state_safe(tracking_store, task_run, TaskRunState.FAILED)
            return


def set_task_run_state_safe(tracking_store, task_run, state):
    # type: (TrackingStore, Any, TaskRunState) -> None
    try:
        tracking_store.set_task_run_state(task_run=task_run, state=state)
    except Exception as exc:
        logger.error("Unable to set task run state: %s", exc)


def save_external_links_safe(tracking_store, task_run, links_dict):
    # type: (TrackingStore, Any, Dict[str, str]) -> None
    try:
        tracking_store.save_external_links(
            task_run=task_run, external_links_dict=links_dict
        )
    except Exception as exc:
        logger.error("Unable to set external links: %s", exc)


@contextmanager
def track_dataproc_pyspark_operator(operator, tracking_info):
    if operator.dataproc_properties is None:
        operator.dataproc_properties = dict()
    spark_envs = get_dbnd_context_spark_conf(
        tracking_info, spark_conf=operator.dataproc_properties
    )

    # It's possible to pass JARs to Dataproc operator in two ways:
    # 1. `spark.jars` property
    # 2. `dataproc_jars` property
    # Those properties are mutually exclude each other
    # So if dataproc_jars are populated we're gonna add our jar here
    if operator.dataproc_jars and len(operator.dataproc_jars) > 0:
        # del spark.jars from spark props because jars are in `jar_file_uris` property already
        operator.dataproc_jars = _merge_jars(
            operator.dataproc_jars, spark_envs.pop("spark.jars", None)
        )

    operator.dataproc_properties.update(spark_envs)
    yield


@contextmanager
def track_dataproc_submit_job_operator(operator, tracking_info):
    logger.info("Tracking DataprocSubmitJobOperator")
    if operator.job is not None:
        if "pyspark_job" in operator.job:
            logger.info("Updating properties for %s", operator)

            pyspark_job = operator.job["pyspark_job"]
            pyspark_job.setdefault("properties", {})

            spark_envs = get_dbnd_context_spark_conf(
                tracking_info=tracking_info, spark_conf=pyspark_job["properties"]
            )

            if "jar_file_uris" in pyspark_job and len(pyspark_job["jar_file_uris"]) > 0:
                # merge `jar_file_uris` property
                pyspark_job["jar_file_uris"] = _merge_jars(
                    pyspark_job["jar_file_uris"], spark_envs.pop("spark.jars", None)
                )

            pyspark_job["properties"].update(spark_envs)
            logger.info(
                "Properties for pyspark_job was updated for task %s", operator.task_id
            )
        else:
            logger.info("pyspark_job has been not found in the operator")
    yield


@contextmanager
def track_spark_submit_operator(operator, tracking_info):
    # type: (SparkSubmitOperator, Dict[str,str])-> None
    if operator._conf is None:
        operator._conf = dict()
    operator._conf = get_dbnd_context_spark_conf(
        tracking_info, spark_conf=operator._conf
    )
    operator._env_vars = get_dbnd_context_env_vars(
        tracking_info, env_vars=operator._env_vars
    )

    yield


@contextmanager
def track_ecs_operator(operator, tracking_info):
    # type: (ECSOperator, Dict[str,str])-> None
    """
    Adding the tracking info to the ECS environment through the `override` -> `containerOverrides`.
    Notice that we require to have `overrides` and `containerOverrides` with containers names in-ordered to make it work

    Airflow pass the override to boto so here is the boto3 docs:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html#ECS.Client.run_task
    """
    info_as_env_var = [
        {"name": key, "value": value} for key, value in six.iteritems(tracking_info)
    ]

    new = []
    if "containerOverrides" in operator.overrides:
        for override in operator.overrides["containerOverrides"]:
            override.setdefault("environment", [])
            override["environment"].extend(info_as_env_var)
            new.append(override)

        operator.overrides["containerOverrides"] = new
    yield


@contextmanager
def track_custom_spark_submit_operator(operator, tracking_info):
    # CHECK that your operator doesn't have any other spark.jars
    # enables DbndSparkQueryExecutionListener. - it can be omitted as a first step

    try:
        # UPDATE CUSTOM OPERATOR
        custom_cli = convert_spark_conf_to_cli_args(
            get_dbnd_context_spark_conf(tracking_info)
        )

        if not hasattr(operator, "spark_args"):
            dbnd_log_exception(
                "User Spark Operator doesn't have attribute 'spark_args', "
                "skipping the context injection"
            )
        else:
            if operator.spark_args is None:
                operator.spark_args = []
            operator.spark_args.extend(custom_cli)

    except Exception as ex:
        dbnd_log_exception("Failed to wrap '%s'. Got an error '%s'", operator, ex)

    yield



# © Copyright Databand.ai, an IBM Company 2022

import logging

import sh

from dbnd_airflow.compat import AIRFLOW_VERSION_1, AIRFLOW_VERSION_2


def set_airflow_connection(
    conn_id, conn_type=None, conn_uri=None, extra=None, host=None
):
    airflow_command = ["connections"]

    def _support_v1(arg):
        if AIRFLOW_VERSION_1:
            return arg[0:2] + arg[2:].replace("-", "_")
        return arg

    # name
    airflow_command.extend(
        ["add", conn_id] if AIRFLOW_VERSION_2 else ["--add", "--conn_id", conn_id]
    )
    # conn type
    if conn_type:
        airflow_command.extend([_support_v1("--conn-type"), conn_type])
    if extra:
        airflow_command.extend([_support_v1("--conn-extra"), extra])
    if conn_uri:
        airflow_command.extend([_support_v1("--conn-uri"), conn_uri])
    if host:
        airflow_command.extend([_support_v1("--conn-host"), host])

    try:
        if AIRFLOW_VERSION_2:
            sh.airflow(["connections", "delete", conn_id])
        else:
            sh.airflow(["connections", "--delete", "--conn_id", conn_id])

    except sh.ErrorReturnCode:
        pass

    logging.info("running: airflow {}".format(" ".join(airflow_command)))
    sh.airflow(airflow_command, _truncate_exc=False)


def run_dag_backfill(dag_id, backfill_date, ok_exit_codes=None) -> int:
    if AIRFLOW_VERSION_2:
        airflow_command = f"dags backfill -s {backfill_date} -e {backfill_date} {dag_id} --reset-dagruns --yes"
    else:
        airflow_command = f"backfill -s {backfill_date} -e {backfill_date} {dag_id} --reset_dagruns --yes"

    logging.info("running: airflow {}".format(airflow_command))
    try:
        airflow_process = sh.airflow(
            airflow_command.split(), _truncate_exc=False, _ok_code=ok_exit_codes or 0
        )
    except Exception:
        logging.exception(
            "something went wrong, dags backfill process exited abnormally"
        )
        return -1

    return airflow_process


def dags_unpause(dag_id):
    if AIRFLOW_VERSION_2:
        sh.airflow(["dags", "unpause", dag_id], _truncate_exc=False)
    else:
        sh.airflow(["unpause", dag_id], _truncate_exc=False)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import uuid

from dbnd._core.utils.basics.memoized import cached
from dbnd._core.utils.uid_utils import get_job_run_uid, get_task_run_attempt_uid


logger = logging.getLogger(__name__)


def get_task_run_attempt_uid_from_af_ti(ti):
    airflow_instance_uid = get_airflow_instance_uid()
    run_uid = get_job_run_uid(
        airflow_instance_uid=airflow_instance_uid,
        dag_id=ti.dag_id,
        execution_date=ti.execution_date,
    )
    return get_task_run_attempt_uid(run_uid, ti.dag_id, ti.task_id, ti.try_number)


@cached()
def get_airflow_instance_uid():
    """used to distinguish between jobs of different airflow instances"""
    import airflow

    db_url = airflow.settings.Session.bind.engine.url
    db_str = "{}:{}/{}".format(db_url.host, db_url.port, db_url.database)
    airflow_instance_uid = uuid.uuid5(uuid.NAMESPACE_URL, db_str)
    return str(airflow_instance_uid)



# © Copyright Databand.ai, an IBM Company 2022

from os import path

import setuptools

from setuptools.config import read_configuration


BASE_PATH = path.dirname(__file__)
CFG_PATH = path.join(BASE_PATH, "setup.cfg")

config = read_configuration(CFG_PATH)
version = config["metadata"]["version"]

setuptools.setup(
    name="dbnd-airflow",
    package_dir={"": "src"},
    install_requires=["dbnd==" + version, "packaging"],
    # Only for orchestration, tracking users should install Airflow manually before DBND
    # The best way to install airflow is manually with constraints beforehand.
    # For example:
    # pip install apache-airflow  --constraint "https://raw.githubusercontent.com/apache/airflow/constraints-${AIRFLOW_VERSION}/constraints-{PYTHON_VERSION}.txt
    extras_require=dict(
        tests=[
            # airflow 2.3 has a problem with pluggy <1.0, that makes pytest 4 incompatible
            "pytest==6.2.5",
            "coverage==7.0.1",
            "pytest-cov==3.0.0",
            "boto3",
            "mock",
            "sh",
        ]
    ),
)



# © Copyright Databand.ai, an IBM Company 2022


import airflow

from dbnd._vendor import version


AIRFLOW_VERSION_BEFORE_2_2 = version.parse(airflow.version.version) < version.parse(
    "2.2.0"
)
AIRFLOW_VERSION_2 = version.parse(airflow.version.version) >= version.parse("2.0.0")

AIRFLOW_VERSION_1 = version.parse(airflow.version.version) < version.parse("2.0.0")

if AIRFLOW_VERSION_1:
    from airflow.hooks.base_hook import BaseHook

elif AIRFLOW_VERSION_2:
    from airflow.hooks.base import BaseHook

else:
    raise NotImplementedError

__all__ = [
    "BaseHook",
    "AIRFLOW_VERSION_1",
    "AIRFLOW_VERSION_2",
    "AIRFLOW_VERSION_BEFORE_2_2",
]



# © Copyright Databand.ai, an IBM Company 2022

import logging


logger = logging.getLogger(__name__)


def reinit_airflow_sql_conn():
    from airflow.settings import configure_orm, configure_vars

    configure_vars()
    # The webservers import this file from models.py with the default settings.
    configure_orm()
    # add query handler before every execute
    # this will print query, code line and stack trace
